package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.math.TransformationMatrix;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.core.model.core.ImplicitGeometry;

import java.io.Serial;

/**
 * Datastructure for representation and resolving of implicit geometries
 *
 * @author Riegel
 */
public class ImplicitGeometryHolder extends Geometry {

    private static final Logger logger = LogManager.getLogger(LibraryObject.class);
    @Serial
    private static final long serialVersionUID = -8938931081577196349L;

    private ImplicitGeometry cgmlImplicitGeometry = null;
    /*
        LibraryObject and RelativeGeometry are mutually exclusive
     */
    private final LibraryObject libObject;
    private final RelativeGeometry relativeGeom;

    public ImplicitGeometryHolder(ImplicitGeometry cgmlImplicitGeometry, LibraryObject libObject) {
        super(libObject.getType(), libObject.getLod());
        this.libObject = libObject;
        this.relativeGeom = null;
        this.cgmlImplicitGeometry = cgmlImplicitGeometry;
        applyTransformation();
    }

    public ImplicitGeometryHolder(ImplicitGeometry cgmlImplicitGeometry, RelativeGeometry relativeGeometry) {
        super(relativeGeometry.getType(), relativeGeometry.getLod());
        this.libObject = null;
        this.relativeGeom = relativeGeometry;
        this.cgmlImplicitGeometry = cgmlImplicitGeometry;
        applyTransformation();
    }


    @Override
    public void accept(Check c) {
        /**
         * Checks are only run on the referenced prototypical geometry.
         */
        if (relativeGeom != null) {
            relativeGeom.accept(c);
        }
        if (libObject != null) {
            libObject.accept(c);
        }
    }

    /**
     * Applies the transformation matrix of the implicit geometry to a copy of the reference geometry and copies the resulting
     * transformed geometry into this object.
     */
    private void applyTransformation() {
        TransformationMatrix transformMatrix = new TransformationMatrix(cgmlImplicitGeometry);
        Geometry refGeo = null;
        if (libObject == null) {
            refGeo = relativeGeom;
        } else if (relativeGeom == null) {
            refGeo = libObject;
        }
        if (refGeo == null) {
            throw new NullPointerException("Geometry reference object of ImplicitGeometryHolder is null!");
        }
        Geometry transformedGeom = transformMatrix.transformGeometry(refGeo);
        transformedGeom.getPolygons().forEach(this::addPolygon);
        this.updateEdgesAndVertices();
    }
}
