package de.hft.stuttgart.citydoctor2.gui;

import de.hft.stuttgart.citydoctor2.gui.tree.Renderable;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.*;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.HBox;
import javafx.stage.Stage;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.io.InputStream;

public class MainToolBar {

    private static final Logger logger = LogManager.getLogger(MainToolBar.class);

    @FXML
    private Button saveBtn;

    @FXML
    private ImageView saveView;

    @FXML
    private Button openBtn;

    @FXML
    private ImageView openImageView;

    @FXML
    private Button checkBtn;

    @FXML
    private ImageView checkImageView;

    @FXML
    private Button showWorldBtn;

    @FXML
    private ImageView showWorldImageView;

    @FXML
    private ToggleButton hideRoofBtn;

    @FXML
    private ImageView hideRoofView;

    @FXML
    private Button resetCameraBtn;

    @FXML
    private ImageView resetCameraView;

    @FXML
    private Button schematronBtn;

    @FXML
    private ImageView schematronImgView;

    @FXML
    private ToggleButton lod1Btn;

    @FXML
    private ImageView lod1View;

    @FXML
    private ToggleButton lod2Btn;

    @FXML
    private ImageView lod2View;

    @FXML
    private ToggleButton lod3Btn;

    @FXML
    private ImageView lod3View;

    @FXML
    private ToggleButton lod4Btn;

    @FXML
    private ImageView lod4View;

    @FXML
    private Button aboutBtn;

    @FXML
    private ImageView aboutImgView;

    @FXML
    private ToggleButton gridButton;

    @FXML
    private ToggleButton cullingButton;

    @FXML
    private ImageView cullingImageView;

    @FXML
    private ImageView gridImageView;

    @FXML
    private Button reportBtn;

    @FXML
    private ImageView reportImageView;

    private OpenFileDialog fileDialog;
    private CheckDialog checkDialog;
    private WriteReportDialog writeDialog;
    private AboutDialog aboutDialog;

    private CityDoctorController controller;
    private TabPane featurePane;
    private Stage stage;
    private Renderer renderer;
    private MainWindow mainWindow;

    private HBox toolBar;

    public MainToolBar(Stage stage, CityDoctorController controller, TabPane featurePane, Renderer renderer,
                       MainWindow mainWindow) throws IOException {
        this.controller = controller;
        this.featurePane = featurePane;
        this.renderer = renderer;
        this.stage = stage;
        this.mainWindow = mainWindow;

        FXMLLoader loader = new FXMLLoader(MainToolBar.class.getResource("MainToolBar.fxml"));
        loader.setController(this);
        toolBar = loader.load();

        fileDialog = new OpenFileDialog(stage, controller);
    }

    public void initialize() {
        openBtn.setOnAction(ae -> fileDialog.show());

        setupSaveBtn();
        setupCheckButton();
        setupLodButtons();
        setupAboutButton();
        setupReportButton();
        setupResetCameraButton();
        setupHideRoofButton();
        loadImages();

        gridButton.setOnAction(ae -> renderer.showWireFrame(gridButton.isSelected()));
        gridButton.setTooltip(new Tooltip(Localization.getText("MainToolBar.wireframe")));
        cullingButton.setOnAction(ae -> renderer.enableCulling(cullingButton.isSelected()));
        cullingButton.setTooltip(new Tooltip(Localization.getText("MainToolBar.culling")));

        showWorldBtn.setOnAction(ar -> {
            Tab selectedItem = featurePane.getSelectionModel().getSelectedItem();
            if (selectedItem.getContent() instanceof TreeView) {
                @SuppressWarnings("unchecked")
                TreeView<Renderable> content = (TreeView<Renderable>) selectedItem.getContent();
                content.getSelectionModel().clearSelection();
            }
            controller.showWorld();
        });
        showWorldBtn.setTooltip(new Tooltip(Localization.getText("MainToolBar.showWorld")));
    }

    private void loadImages() {
        try {
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/openFolderIcon.png")) {
                Image img = new Image(inStream);
                openImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/check32x32.png")) {
                Image img = new Image(inStream);
                checkImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/wireframe32x32.png")) {
                Image img = new Image(inStream);
                gridImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/Culling.png")) {
                Image img = new Image(inStream);
                cullingImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/error_stat32x32.png")) {
                Image img = new Image(inStream);
                reportImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/scene.png")) {
                Image img = new Image(inStream);
                showWorldImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/lod1_32x32.png")) {
                Image img = new Image(inStream);
                lod1View.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/lod2_32x32.png")) {
                Image img = new Image(inStream);
                lod2View.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/lod3_32x32.png")) {
                Image img = new Image(inStream);
                lod3View.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/lod4_32x32.png")) {
                Image img = new Image(inStream);
                lod4View.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/about.png")) {
                Image img = new Image(inStream);
                aboutImgView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/save.png")) {
                Image img = new Image(inStream);
                saveView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/reset_camera.png")) {
                Image img = new Image(inStream);
                resetCameraView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/hide_roof.png")) {
                Image img = new Image(inStream);
                hideRoofView.setImage(img);
            }
        } catch (IOException e) {
            // ignore close exception
        }
    }

    private void setupReportButton() {
        reportBtn.setDisable(true);
        reportBtn.setTooltip(new Tooltip(Localization.getText("MainToolBar.writeReports")));
        reportBtn.setOnAction(ae -> {
            if (writeDialog == null) {
                try {
                    writeDialog = new WriteReportDialog(stage, controller, mainWindow);
                } catch (IOException e) {
                    logger.catching(e);
                    mainWindow.showExceptionDialog(e);
                }
            }
            if (writeDialog != null) {
                // writeDialog can be null if creation of said dialog fails
                writeDialog.show();
            }
        });
    }

    private void setupAboutButton() {
        aboutBtn.setOnAction(ae -> {
            if (aboutDialog == null) {
                try {
                    aboutDialog = new AboutDialog(stage);
                    aboutDialog.show();
                } catch (IOException e) {
                    logger.error("Could not load about dialog.", e);
                }
            } else {
                aboutDialog.show();
            }
        });
    }

    private void setupLodButtons() {
        lod1Btn.setOnAction(ae -> {
            if (lod1Btn.isSelected()) {
                renderer.enableLod1();
            } else {
                renderer.disableLod1();
            }
        });
        lod2Btn.setOnAction(ae -> {
            if (lod2Btn.isSelected()) {
                renderer.enableLod2();
            } else {
                renderer.disableLod2();
            }
        });
        lod3Btn.setOnAction(ae -> {
            if (lod3Btn.isSelected()) {
                renderer.enableLod3();
            } else {
                renderer.disableLod3();
            }
        });
        lod4Btn.setOnAction(ae -> {
            if (lod4Btn.isSelected()) {
                renderer.enableLod4();
            } else {
                renderer.disableLod4();
            }
        });
    }

    private void setupSaveBtn() {
        saveBtn.setOnAction(ae -> controller.askAndSave());
    }

    private void setupCheckButton() {
        checkBtn.setDisable(true);
        checkBtn.setTooltip(new Tooltip(Localization.getText("MainToolBar.executeChecks")));
        checkBtn.setOnAction(ae -> {
            if (checkDialog == null) {
                try {
                    checkDialog = new CheckDialog(mainWindow, stage, controller);
                } catch (IOException e) {
                    mainWindow.showExceptionDialog(e);
                    logger.catching(e);
                }
            }
            checkDialog.show();
        });
    }

    private void setupResetCameraButton() {
        resetCameraBtn.setDisable(true);
        resetCameraBtn.setTooltip(new Tooltip(Localization.getText("MainToolBar.resetCamera")));
        resetCameraBtn.setOnAction(ae -> mainWindow.resetCamera());

    }

    private void setupHideRoofButton() {
        hideRoofBtn.setDisable(true);
        hideRoofBtn.setSelected(false);
        hideRoofBtn.setTooltip(new Tooltip(Localization.getText("MainToolBar.hideRoof")));
        hideRoofBtn.setOnAction(ae -> {
            if (hideRoofBtn.isSelected()) {
                renderer.hideRoofs();
            } else {
                renderer.showRoofs();
            }
        });

    }

    public Button getCheckButton() {
        return checkBtn;
    }

    public ToggleButton getGridButton() {
        return gridButton;
    }

    public ToggleButton getCullingButton() {
        return cullingButton;
    }

    public Button getSaveButton() {
        return saveBtn;
    }

    public Button getWriteReportButton() {
        return reportBtn;
    }

    public ToggleButton getLod1Btn() {
        return lod1Btn;
    }

    public ToggleButton getLod2Btn() {
        return lod2Btn;
    }

    public ToggleButton getLod3Btn() {
        return lod3Btn;
    }

    public ToggleButton getLod4Btn() {
        return lod4Btn;
    }

    public ToggleButton getHideRoofBtn() {
        return hideRoofBtn;
    }

    public Button getWorldBtn() {
        return showWorldBtn;
    }

    public Button getSaveBtn() {
        return saveBtn;
    }

    public Button getResetCameraBtn() {
        return resetCameraBtn;
    }

    public HBox getToolBar() {
        return toolBar;
    }

    public Button getOpenBtn() {
        return openBtn;
    }

}
