package de.hft.stuttgart.citydoctor2.healing;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;

import org.citygml4j.xml.writer.CityGMLWriteException;
import org.junit.Assert;
import org.junit.Test;

import de.hft.stuttgart.citydoctor2.CppInitializer;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.check.error.SolidNotClosedError;
import de.hft.stuttgart.citydoctor2.checks.geometry.SolidNotClosedCheck;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.healer.Healer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;

public class HealSolidNotClosedErrorTest {

	/**
	 * One polygon is split up in 3 but has not matching edges
	 */
	@Test
	public void testVertexCloseToEdge() {
		CppInitializer.initCpp();
		Building b = new Building(); //Angelika: als Test eingefügt
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD0); //Original
//		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD2); //Angelika: Test
		b.addGeometry(geom); //Angelika: als Test eingefügt
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v2 = new Vertex(4, 2, 0);
		Vertex v3 = new Vertex(2, 2, 0);
		Vertex v4 = new Vertex(4, 4, 0);
		Vertex v5 = new Vertex(2, 4, 0);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v7 = new Vertex(0, 2, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 4);
		
		// error polygons
		Polygon p1 = new ConcretePolygon();
		geom.addPolygon(p1);
		LinearRing r1 = new LinearRing(LinearRingType.EXTERIOR);
		p1.setExteriorRing(r1);
		r1.addVertex(v0);
		r1.addVertex(v1);
		r1.addVertex(v2);
		r1.addVertex(v7);
		r1.addVertex(v0);
		
		Polygon p2 = new ConcretePolygon();
		geom.addPolygon(p2);
		LinearRing r2 = new LinearRing(LinearRingType.EXTERIOR);
		p2.setExteriorRing(r2);
		r2.addVertex(v2);
		r2.addVertex(v4);
		r2.addVertex(v5);
		r2.addVertex(v3);
		r2.addVertex(v2);
		
		Polygon p3 = new ConcretePolygon();
		geom.addPolygon(p3);
		LinearRing r3 = new LinearRing(LinearRingType.EXTERIOR);
		p3.setExteriorRing(r3);
		r3.addVertex(v7);
		r3.addVertex(v3);
		r3.addVertex(v5);
		r3.addVertex(v6);
		r3.addVertex(v7);
		
		// close the shell
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p7 = new ConcretePolygon();
		geom.addPolygon(p7);
		LinearRing r7 = new LinearRing(LinearRingType.EXTERIOR);
		p7.setExteriorRing(r7);
		r7.addVertex(v4);
		r7.addVertex(v9);
		r7.addVertex(v10);
		r7.addVertex(v6);
		r7.addVertex(v4);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);
		r8.addVertex(v0);
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v0);
		
		geom.updateEdgesAndVertices();

		SolidNotClosedCheck check = new SolidNotClosedCheck();
		check.check(geom);
		CheckResult cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_S_NOT_CLOSED, cr.getError().getErrorId());
		assertTrue(cr.getError() instanceof SolidNotClosedError);
		
//		HealSolidNotClosedError healer = new HealSolidNotClosedError(); //Original
		HealSolidNotClosedCpp healer = new HealSolidNotClosedCpp(); // Angelika
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();
		check.check(geom);
		cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.OK, cr.getResultStatus());
	}
	
	@Test
	public void testMissingPolygon() {
		CppInitializer.initCpp();
		Building b = new Building();
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD2);
		b.addGeometry(geom);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v4 = new Vertex(4, 4, 0);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 4);
		
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p7 = new ConcretePolygon();
		geom.addPolygon(p7);
		LinearRing r7 = new LinearRing(LinearRingType.EXTERIOR);
		p7.setExteriorRing(r7);
		r7.addVertex(v4);
		r7.addVertex(v9);
		r7.addVertex(v10);
		r7.addVertex(v6);
		r7.addVertex(v4);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);
		r8.addVertex(v0);
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v0);
		
		geom.updateEdgesAndVertices();
		
		SolidNotClosedCheck check = new SolidNotClosedCheck();
		check.check(geom);
		CheckResult cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_S_NOT_CLOSED, cr.getError().getErrorId());
		assertTrue(cr.getError() instanceof SolidNotClosedError);
		
		HealSolidNotClosedCpp healer = new HealSolidNotClosedCpp();
//		HealSolidNotClosedError healer = new HealSolidNotClosedError(); //Angelika: Test
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();
		geom.clearAllContainedCheckResults();
		check.check(geom);
		cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.OK, cr.getResultStatus());
		
		
	}
	
	@Test
	public void test2MissingPolygons() {
		Building b = new Building(); 
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD0);
		b.addGeometry(geom); 
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v4 = new Vertex(4, 4, 0);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 4);
		
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);
		r8.addVertex(v0);
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v0);
		
		geom.updateEdgesAndVertices();
		
		SolidNotClosedCheck check = new SolidNotClosedCheck();
		check.check(geom);
		CheckResult cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_S_NOT_CLOSED, cr.getError().getErrorId());
		assertTrue(cr.getError() instanceof SolidNotClosedError);
		
		HealSolidNotClosedError healer = new HealSolidNotClosedError();
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();
		
		geom.clearAllContainedCheckResults();
		check.check(geom);
		cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_S_NOT_CLOSED, cr.getError().getErrorId());
		assertTrue(cr.getError() instanceof SolidNotClosedError);

		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();
		
		geom.clearAllContainedCheckResults();
		check.check(geom);
		cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.OK, cr.getResultStatus());
	}
	
	@Test
	public void test2MissingPolygonsCpp() {
		CppInitializer.initCpp();
		Building b = new Building(); 
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD2); 
		b.addGeometry(geom);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v4 = new Vertex(4, 4, 0);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 4);
		
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);
		r8.addVertex(v0);
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v0);
		
		geom.updateEdgesAndVertices();
		
		SolidNotClosedCheck check = new SolidNotClosedCheck();
		check.check(geom);
		CheckResult cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_S_NOT_CLOSED, cr.getError().getErrorId());
		assertTrue(cr.getError() instanceof SolidNotClosedError);
		
//		HealSolidNotClosedError healer = new HealSolidNotClosedError(); //Original
		HealSolidNotClosedCpp healer = new HealSolidNotClosedCpp(); // Angelika
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();
		
		geom.clearAllContainedCheckResults();
		check.check(geom);
		cr = geom.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.OK, cr.getResultStatus());
	}
	
	/**
	 * Missing Polygon has a self intersection
	 * @throws IOException 
	 * @throws CityGmlParseException 
	 * @throws CityGMLWriteException 
	 * @throws CityGMLBuilderException 
	 * @throws InvalidGmlFileException 
	 */
	@Test
	public void testMissingPolygon2() throws CityGmlParseException, IOException, InvalidGmlFileException {
		String path = "src/test/resources/SimpleSolid_SrefBS-GE-gml-SO-0002-T0002.gml"; //bei diesem Modell stürzt die VM ab
//		String path = "src/test/resources/SimpleSolid_SrefBS-GE-gml-SO-0002-T0001.gml"; //hier nur Fehlermeldung
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		geom.updateVertices();
		for (Vertex v : geom.getVertices()) {
			v.setX(v.getX() - 10);
			v.setY(v.getY() - 10);
		}
		Assert.assertTrue(geom.containsError(CheckId.C_GE_S_NOT_CLOSED));
		
		Healer healer = new Healer(c);
		//healer.setNumberOfIterations(2);
		healer.healCityObject(model.getBuildings().get(0));
		
		Assert.assertFalse(geom.containsError(CheckId.C_GE_S_NOT_CLOSED));
	}

}
